/* (C) 2022 YiRing, Inc. */
package com.yiring.app.domain.location;

import com.yiring.auth.domain.user.User;
import com.yiring.common.annotation.FieldMapping;
import com.yiring.common.domain.BasicEntity;
import java.io.Serial;
import java.io.Serializable;
import java.time.LocalDateTime;
import java.util.Objects;
import javax.persistence.*;
import lombok.*;
import lombok.experimental.FieldDefaults;
import lombok.experimental.FieldNameConstants;
import lombok.experimental.SuperBuilder;
import org.hibernate.Hibernate;
import org.hibernate.annotations.Comment;

/**
 * 定位标签
 * 引用:
 *  1. 定位平台接口规范V3.0.1 #4.3
 *  2. 真源定位系统后台管理模块
 *
 * @author Jim
 * @version 0.1
 * 2022/4/7 11:02
 */

@Getter
@Setter
@ToString
@SuperBuilder(toBuilder = true)
@NoArgsConstructor
@AllArgsConstructor
@FieldNameConstants
@FieldDefaults(level = AccessLevel.PRIVATE)
@Entity
@Table(
    name = "BS_LOCATION_TAG",
    indexes = {
        @Index(columnList = "linkId"),
        @Index(columnList = "type"),
        @Index(columnList = "used"),
        @Index(columnList = "user_id"),
        @Index(columnList = "code", unique = true),
    }
)
@Comment("定位标签")
public class LocationTag extends BasicEntity implements Serializable {

    @Serial
    private static final long serialVersionUID = 5419734189897829250L;

    /**
     * 数据来源于【真源人员定位系统 - 定位标签】
     * 作用: 用于双向联动进行数据同步
     */
    @FieldMapping("id")
    @Comment("外链主键")
    Long linkId;

    @FieldMapping
    @Comment("IMEI 设备标识码")
    String imei;

    /**
     * 前缀: BTT
     * 后缀: 8 位数字
     */
    @FieldMapping("tagId")
    @Comment("编号")
    @Column(unique = true, nullable = false)
    String code;

    @FieldMapping("tagCode")
    @Comment("型号")
    @Enumerated(EnumType.STRING)
    Type type;

    @FieldMapping
    @Comment("静止/运动")
    Boolean silent;

    @OneToOne
    @JoinColumn(name = "user_id")
    @Comment("绑定用户")
    User user;

    @FieldMapping
    @Comment("使用/闲置")
    Boolean used;

    @FieldMapping
    @Comment("电量")
    Integer volt;

    @FieldMapping
    @Comment("电量单位")
    String voltUnit;

    @FieldMapping
    @Comment("类型(1:内部/2:访客/3:承包商)")
    Integer category;

    @FieldMapping(value = "raiseTimestamp", desc = "更新时间戳", type = Long.class)
    @Comment("更新时间")
    LocalDateTime updateTime;

    @Comment("创建时间")
    LocalDateTime createTime;

    @SuppressWarnings({ "unused" })
    public enum Type {
        BTT01("蓝牙人员定位卡"),
        BTT02("蓝牙车辆定位器");

        final String text;

        Type(String text) {
            this.text = text;
        }

        public String text() {
            return this.text;
        }
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || Hibernate.getClass(this) != Hibernate.getClass(o)) return false;
        LocationTag that = (LocationTag) o;
        return getId() != null && Objects.equals(getId(), that.getId());
    }

    @Override
    public int hashCode() {
        return getClass().hashCode();
    }
}
