/* (C) 2022 YiRing, Inc. */
package com.yiring.app.rabbit.receiver;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONObject;
import com.rabbitmq.client.Channel;
import com.yiring.app.domain.location.*;
import com.yiring.app.domain.log.ZyRealtimeLog;
import com.yiring.app.domain.log.ZyRealtimeLogRepository;
import com.yiring.app.rabbit.config.ZyRabbitConfig;
import com.yiring.app.util.GeoUtils;
import java.io.IOException;
import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.util.*;
import java.util.stream.Collectors;
import javax.annotation.Resource;
import lombok.extern.slf4j.Slf4j;
import org.locationtech.jts.geom.Point;
import org.springframework.amqp.core.Message;
import org.springframework.amqp.rabbit.annotation.RabbitHandler;
import org.springframework.amqp.rabbit.annotation.RabbitListener;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.context.annotation.Configuration;
import org.springframework.data.domain.Example;
import org.springframework.messaging.simp.SimpMessagingTemplate;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;

/**
 * 真源 RabbitMQ 消息监听（消息消费者）
 *
 * @author Jim
 * @version 0.1
 * 2022/4/13 17:13
 */

@Slf4j
@Component
@Configuration
@ConditionalOnProperty(prefix = "zy-config.rabbitmq", value = "enabled", havingValue = "true")
public class ZyRabbitReceiver {

    // TODO
    // 1. 新增消息订阅定时任务，检查是否正常订阅了真源定位系统的消息
    // 2. 如果没有所需的订阅记录，则发起消息订阅（见接口文档: 6.2）
    // 3. 订阅 position（定位数据）、lowPower（低电量报警）、deviceStatus（设备状态）、keyWarning（按键报警）

    @Resource
    LocationTagRepository locationTagRepository;

    @Resource
    LocationBeaconRepository locationBeaconRepository;

    @Resource
    LocationLogRepository locationLogRepository;

    @Resource
    SimpMessagingTemplate simpMessagingTemplate;

    @Resource
    ZyRealtimeLogRepository zyRealtimeLogRepository;

    /**
     * 订阅真源定位系统 RabbitMQ 推送过来的消息（主动订阅的一些消息类别）
     * 参见: 定位平台接口规范V3.0.1 - 通用版.pdf #6
     * @param msg 消息内容
     * @param channel 消息通道
     * @param message 消息主体
     * @throws IOException 消息确认异常
     */
    @RabbitHandler
    @RabbitListener(
        queues = ZyRabbitConfig.MESSAGE_QUEUES_NAME,
        containerFactory = ZyRabbitConfig.LISTENER_FACTORY_NAME
    )
    public void listen(String msg, Channel channel, Message message) throws IOException {
        // 手动确认消息已收到
        channel.basicAck(message.getMessageProperties().getDeliveryTag(), false);

        try {
            // 将消息转换成 JSON 格式
            JSONObject info = JSON.parseObject(msg);
            log.info("Receiver Message: {}", info);

            // 解构消息内容
            JSONObject data = info.getJSONObject("params");
            String method = info.getString("method");

            // 记录日志
            ZyRealtimeLog realtimeLog = ZyRealtimeLog.builder().method(method).raw(info).build();
            zyRealtimeLogRepository.save(realtimeLog);

            // 业务处理
            switch (method) {
                // 实时定位
                case "position" -> processPositionMessage(data);
                // 设备低电量
                case "lowPower" -> processLowPowerMessage(data);
                // 设备状态变更
                case "deviceStatus" -> processDeviceStatusMessage(data);
                // 按键报警
                case "keyWarning" -> processKeyWarningMessage(data);
                // 围栏报警
                case "enclosure" -> log.warn("Ignore Message Type [enclosure]: {}", info);
                default -> log.warn("Unknown Message Type: {}", info);
            }
        } catch (Exception e) {
            log.error(e.getMessage(), e);
        }
    }

    /**
     * 处理定位消息
     * @param data 消息内容
     */
    @Transactional(rollbackFor = Exception.class)
    public void processPositionMessage(JSONObject data) {
        // TODO
        log.info("Position Message: {}", data);

        // 包装消息
        // TODO
        // 1. 解析消息内容，进行围栏、出入标识判断等处理，将定位记录录入数据库
        // 2. 创建一条需要进行消息推送的记录
        // 3. 将记录推送的消息推送模块
        // 4. 检查是否触发围栏告警，记录告警数据，并推送消息

        // 定位时间
        Instant instant = Instant.ofEpochMilli(data.getLongValue("locationTime"));
        LocalDateTime time = LocalDateTime.ofInstant(instant, ZoneId.systemDefault());

        // 定位的基本信息
        LocationLog locationLog = LocationLog
            .builder()
            .raw(data)
            .locationTime(time)
            .areaId(data.getLong("areaId"))
            .silent(data.getBoolean("silent"))
            .volt(data.getInteger("volt"))
            .voltUnit(data.getString("voltUnit"))
            .build();

        // 设置空间点位信息
        Point point = GeoUtils.createPoint(
            data.getDoubleValue("longitude"),
            data.getDoubleValue("latitude"),
            data.getDoubleValue("altitude")
        );
        locationLog.setPoint(point);

        // 查询定位标签
        Example<LocationTag> example = Example.of(LocationTag.builder().code(data.getString("tagId")).build());
        Optional<LocationTag> optional = locationTagRepository.findOne(example);
        if (optional.isPresent()) {
            LocationTag tag = optional.get();
            // 设置定位标签
            locationLog.setTag(tag);
            // 定位标签当时所属的用户
            locationLog.setUser(tag.getUser());
            // 查询当前用户的状态
            // TODO
        }

        // 查询定位信标
        Set<String> codes = Arrays
            .stream(data.getString("beacons").split(","))
            .map(beacon -> beacon.replaceAll("\\(.*\\)", ""))
            .collect(Collectors.toSet());
        Set<LocationBeacon> beacons = locationBeaconRepository.findByCodeIn(codes);
        locationLog.setBeacons(beacons);

        // 查询定位所在围栏信息
        Set<LocationFence> fences = locationLog
            .getBeacons()
            .stream()
            .map(LocationBeacon::getFences)
            .flatMap(Set::stream)
            .collect(Collectors.toSet());
        locationLog.setFences(fences);

        // TODO
        // 并计算出入标记（围栏、区域）

        // 写入数据
        locationLogRepository.saveAndFlush(locationLog);

        // 更新定位标签卡电量信息
        // TODO

        // WebSocket 推送定位消息
        // 消息内容需要确定 TODO
        simpMessagingTemplate.convertAndSend("/topic/position", "{}");
        // TODO
        // 判断围栏告警是否触发，触发写入告警记录，并推送消息
    }

    /**
     * 处理低电量报警消息
     * @param data 消息内容
     */
    @Transactional(rollbackFor = Exception.class)
    public void processLowPowerMessage(JSONObject data) {
        // TODO
        log.info("LowPower Message: {}", data);
    }

    /**
     * 处理设备状态更新消息
     * @param data 消息内容
     */
    @Transactional(rollbackFor = Exception.class)
    public void processDeviceStatusMessage(JSONObject data) {
        // TODO
        log.info("DeviceStatus Message: {}", data);
    }

    /**
     * 处理按键报警消息
     * @param data 消息内容
     */
    @Transactional(rollbackFor = Exception.class)
    public void processKeyWarningMessage(JSONObject data) {
        // TODO
        log.info("KeyWarning Message: {}", data);
    }
}
